# -*- coding: utf-8 -*-
"""Format data into a vertical table layout."""

from __future__ import unicode_literals

from cli_helpers.utils import filter_dict_by_key
from .preprocessors import convert_to_string, override_missing_value, style_output

supported_formats = ("vertical",)
preprocessors = (override_missing_value, convert_to_string, style_output)


def _get_separator(num, sep_title, sep_character, sep_length):
    """Get a row separator for row *num*."""
    left_divider_length = right_divider_length = sep_length
    if isinstance(sep_length, tuple):
        left_divider_length, right_divider_length = sep_length
    left_divider = sep_character * left_divider_length
    right_divider = sep_character * right_divider_length
    title = sep_title.format(n=num + 1)

    return "{left_divider}[ {title} ]{right_divider}\n".format(
        left_divider=left_divider, right_divider=right_divider, title=title
    )


def _format_row(headers, row):
    """Format a row."""
    formatted_row = [" | ".join(field) for field in zip(headers, row)]
    return "\n".join(formatted_row)


def vertical_table(
    data, headers, sep_title="{n}. row", sep_character="*", sep_length=27
):
    """Format *data* and *headers* as an vertical table.

    The values in *data* and *headers* must be strings.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param str sep_title: The title given to each row separator. Defaults to
                          ``'{n}. row'``. Any instance of ``'{n}'`` is
                          replaced by the record number.
    :param str sep_character: The character used to separate rows. Defaults to
                              ``'*'``.
    :param int/tuple sep_length: The number of separator characters that should
                                 appear on each side of the *sep_title*. Use
                                 a tuple to specify the left and right values
                                 separately.
    :return: The formatted data.
    :rtype: str

    """
    header_len = max([len(x) for x in headers])
    padded_headers = [x.ljust(header_len) for x in headers]
    formatted_rows = [_format_row(padded_headers, row) for row in data]

    output = []
    for i, result in enumerate(formatted_rows):
        yield _get_separator(i, sep_title, sep_character, sep_length) + result


def adapter(data, headers, **kwargs):
    """Wrap vertical table in a function for TabularOutputFormatter."""
    keys = ("sep_title", "sep_character", "sep_length")
    return vertical_table(data, headers, **filter_dict_by_key(kwargs, keys))



# -*- coding: utf-8 -*-
"""These preprocessor functions are used to process data prior to output."""

import string
from datetime import datetime

from cli_helpers import utils
from cli_helpers.compat import text_type, int_types, float_types, HAS_PYGMENTS, Token


def truncate_string(
    data, headers, max_field_width=None, skip_multiline_string=True, **_
):
    """Truncate very long strings. Only needed for tabular
    representation, because trying to tabulate very long data
    is problematic in terms of performance, and does not make any
    sense visually.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param int max_field_width: Width to truncate field for display
    :return: The processed data and headers.
    :rtype: tuple
    """
    return (
        (
            [
                utils.truncate_string(v, max_field_width, skip_multiline_string)
                for v in row
            ]
            for row in data
        ),
        [
            utils.truncate_string(h, max_field_width, skip_multiline_string)
            for h in headers
        ],
    )


def convert_to_string(data, headers, **_):
    """Convert all *data* and *headers* to strings.

    Binary data that cannot be decoded is converted to a hexadecimal
    representation via :func:`binascii.hexlify`.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :return: The processed data and headers.
    :rtype: tuple

    """
    return (
        ([utils.to_string(v) for v in row] for row in data),
        [utils.to_string(h) for h in headers],
    )


def convert_to_undecoded_string(data, headers, **_):
    """Convert all *data* and *headers* to hex, if needed.

    Binary data is converted to a hexadecimal representation via
    :func:`binascii.hexlify`.

    Unlike convert_to_string(), None values are left as Nones.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :return: The processed data and headers.
    :rtype: tuple

    """
    return (
        ([utils.to_undecoded_string(v) for v in row] for row in data),
        [utils.to_undecoded_string(h) for h in headers],
    )


def override_missing_value(
    data,
    headers,
    style=None,
    missing_value_token=Token.Output.Null,
    missing_value="",
    **_,
):
    """Override missing values in the *data* with *missing_value*.

    A missing value is any value that is :data:`None`.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param style: Style for missing_value.
    :param missing_value_token: The Pygments token used for missing data.
    :param missing_value: The default value to use for missing data.
    :return: The processed data and headers.
    :rtype: tuple

    """

    def fields():
        for row in data:
            processed = []
            for field in row:
                if field is None and style and HAS_PYGMENTS:
                    styled = utils.style_field(
                        missing_value_token, missing_value, style
                    )
                    processed.append(styled)
                elif field is None:
                    processed.append(missing_value)
                else:
                    processed.append(field)
            yield processed

    return (fields(), headers)


def override_tab_value(data, headers, new_value="    ", **_):
    """Override tab values in the *data* with *new_value*.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param new_value: The new value to use for tab.
    :return: The processed data and headers.
    :rtype: tuple

    """
    return (
        (
            [v.replace("\t", new_value) if isinstance(v, text_type) else v for v in row]
            for row in data
        ),
        headers,
    )


def escape_newlines(data, headers, **_):
    """Escape newline characters (\n -> \\n, \r -> \\r)

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :return: The processed data and headers.
    :rtype: tuple

    """
    return (
        (
            [
                (
                    v.replace("\r", r"\r").replace("\n", r"\n")
                    if isinstance(v, text_type)
                    else v
                )
                for v in row
            ]
            for row in data
        ),
        headers,
    )


def bytes_to_string(data, headers, **_):
    """Convert all *data* and *headers* bytes to strings.

    Binary data that cannot be decoded is converted to a hexadecimal
    representation via :func:`binascii.hexlify`.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :return: The processed data and headers.
    :rtype: tuple

    """
    return (
        ([utils.bytes_to_string(v) for v in row] for row in data),
        [utils.bytes_to_string(h) for h in headers],
    )


def align_decimals(data, headers, column_types=(), **_):
    """Align numbers in *data* on their decimal points.

    Whitespace padding is added before a number so that all numbers in a
    column are aligned.

    Outputting data before aligning the decimals::

        1
        2.1
        10.59

    Outputting data after aligning the decimals::

         1
         2.1
        10.59

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param iterable column_types: The columns' type objects (e.g. int or float).
    :return: The processed data and headers.
    :rtype: tuple

    """
    pointpos = len(headers) * [0]
    data = list(data)
    for row in data:
        for i, v in enumerate(row):
            if column_types[i] is float and type(v) in float_types:
                v = text_type(v)
                pointpos[i] = max(utils.intlen(v), pointpos[i])

    def results(data):
        for row in data:
            result = []
            for i, v in enumerate(row):
                if column_types[i] is float and type(v) in float_types:
                    v = text_type(v)
                    result.append((pointpos[i] - utils.intlen(v)) * " " + v)
                else:
                    result.append(v)
            yield result

    return results(data), headers


def quote_whitespaces(data, headers, quotestyle="'", **_):
    """Quote leading/trailing whitespace in *data*.

    When outputing data with leading or trailing whitespace, it can be useful
    to put quotation marks around the value so the whitespace is more
    apparent. If one value in a column needs quoted, then all values in that
    column are quoted to keep things consistent.

    .. NOTE::
       :data:`string.whitespace` is used to determine which characters are
       whitespace.

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param str quotestyle: The quotation mark to use (defaults to ``'``).
    :return: The processed data and headers.
    :rtype: tuple

    """
    whitespace = tuple(string.whitespace)
    quote = len(headers) * [False]
    data = list(data)
    for row in data:
        for i, v in enumerate(row):
            v = text_type(v)
            if v.startswith(whitespace) or v.endswith(whitespace):
                quote[i] = True

    def results(data):
        for row in data:
            result = []
            for i, v in enumerate(row):
                quotation = quotestyle if quote[i] else ""
                result.append(
                    "{quotestyle}{value}{quotestyle}".format(
                        quotestyle=quotation, value=v
                    )
                )
            yield result

    return results(data), headers


def style_output(
    data,
    headers,
    style=None,
    header_token=Token.Output.Header,
    odd_row_token=Token.Output.OddRow,
    even_row_token=Token.Output.EvenRow,
    **_,
):
    """Style the *data* and *headers* (e.g. bold, italic, and colors)

    .. NOTE::
        This requires the `Pygments <http://pygments.org/>`_ library to
        be installed. You can install it with CLI Helpers as an extra::
            $ pip install cli_helpers[styles]

    Example usage::

        from cli_helpers.tabular_output.preprocessors import style_output
        from pygments.style import Style
        from pygments.token import Token

        class YourStyle(Style):
            default_style = ""
            styles = {
                Token.Output.Header: 'bold ansibrightred',
                Token.Output.OddRow: 'bg:#eee #111',
                Token.Output.EvenRow: '#0f0'
            }

        headers = ('First Name', 'Last Name')
        data = [['Fred', 'Roberts'], ['George', 'Smith']]

        data, headers = style_output(data, headers, style=YourStyle)

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param str/pygments.style.Style style: A Pygments style. You can `create
        your own styles <https://pygments.org/docs/styles#creating-own-styles>`_.
    :param str header_token: The token type to be used for the headers.
    :param str odd_row_token: The token type to be used for odd rows.
    :param str even_row_token: The token type to be used for even rows.
    :return: The styled data and headers.
    :rtype: tuple

    """
    from cli_helpers.utils import filter_style_table

    relevant_styles = filter_style_table(
        style, header_token, odd_row_token, even_row_token
    )
    if style and HAS_PYGMENTS:
        if relevant_styles.get(header_token):
            headers = [
                utils.style_field(header_token, header, style) for header in headers
            ]
        if relevant_styles.get(odd_row_token) or relevant_styles.get(even_row_token):
            data = (
                [
                    utils.style_field(
                        odd_row_token if i % 2 else even_row_token, f, style
                    )
                    for f in r
                ]
                for i, r in enumerate(data, 1)
            )

    return iter(data), headers


def format_numbers(
    data, headers, column_types=(), integer_format=None, float_format=None, **_
):
    """Format numbers according to a format specification.

    This uses Python's format specification to format numbers of the following
    types: :class:`int`, :class:`py2:long` (Python 2), :class:`float`, and
    :class:`~decimal.Decimal`. See the :ref:`python:formatspec` for more
    information about the format strings.

    .. NOTE::
       A column is only formatted if all of its values are the same type
       (except for :data:`None`).

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param iterable column_types: The columns' type objects (e.g. int or float).
    :param str integer_format: The format string to use for integer columns.
    :param str float_format: The format string to use for float columns.
    :return: The processed data and headers.
    :rtype: tuple

    """
    if (integer_format is None and float_format is None) or not column_types:
        return iter(data), headers

    def _format_number(field, column_type):
        if integer_format and column_type is int and type(field) in int_types:
            return format(field, integer_format)
        elif float_format and column_type is float and type(field) in float_types:
            return format(field, float_format)
        return field

    data = (
        [_format_number(v, column_types[i]) for i, v in enumerate(row)] for row in data
    )
    return data, headers


def format_timestamps(data, headers, column_date_formats=None, **_):
    """Format timestamps according to user preference.

    This allows for per-column formatting for date, time, or datetime like data.

    Add a `column_date_formats` section to your config file with separate lines for each column
    that you'd like to specify a format using `name=format`. Use standard Python strftime
    formatting strings

    Example: `signup_date = "%Y-%m-%d"`

    :param iterable data: An :term:`iterable` (e.g. list) of rows.
    :param iterable headers: The column headers.
    :param str column_date_format: The format strings to use for specific columns.
    :return: The processed data and headers.
    :rtype: tuple

    """
    if column_date_formats is None:
        return iter(data), headers

    def _format_timestamp(value, name, column_date_formats):
        if name not in column_date_formats:
            return value
        try:
            dt = datetime.fromisoformat(value)
            return dt.strftime(column_date_formats[name])
        except (ValueError, TypeError):
            # not a date
            return value

    data = (
        [
            _format_timestamp(v, headers[i], column_date_formats)
            for i, v in enumerate(row)
        ]
        for row in data
    )
    return data, headers



# -*- coding: utf-8 -*-
"""Read and write an application's config files."""

from __future__ import unicode_literals
import io
import logging
import os

from configobj import ConfigObj, ConfigObjError
from validate import ValidateError, Validator

from .compat import MAC, text_type, UserDict, WIN

logger = logging.getLogger(__name__)


class ConfigError(Exception):
    """Base class for exceptions in this module."""

    pass


class DefaultConfigValidationError(ConfigError):
    """Indicates the default config file did not validate correctly."""

    pass


class Config(UserDict, object):
    """Config reader/writer class.

    :param str app_name: The application's name.
    :param str app_author: The application author/organization.
    :param str filename: The config filename to look for (e.g. ``config``).
    :param dict/str default: The default config values or absolute path to
                             config file.
    :param bool validate: Whether or not to validate the config file.
    :param bool write_default: Whether or not to write the default config
                               file to the user config directory if it doesn't
                               already exist.
    :param tuple additional_dirs: Additional directories to check for a config
                                  file.
    """

    def __init__(
        self,
        app_name,
        app_author,
        filename,
        default=None,
        validate=False,
        write_default=False,
        additional_dirs=(),
    ):
        super(Config, self).__init__()
        #: The :class:`ConfigObj` instance.
        self.data = ConfigObj(encoding="utf8")

        self.default = {}
        self.default_file = self.default_config = None
        self.config_filenames = []

        self.app_name, self.app_author = app_name, app_author
        self.filename = filename
        self.write_default = write_default
        self.validate = validate
        self.additional_dirs = additional_dirs

        if isinstance(default, dict):
            self.default = default
            self.update(default)
        elif isinstance(default, text_type):
            self.default_file = default
        elif default is not None:
            raise TypeError(
                '"default" must be a dict or {}, not {}'.format(
                    text_type.__name__, type(default)
                )
            )

        if self.write_default and not self.default_file:
            raise ValueError(
                'Cannot use "write_default" without specifying ' "a default file."
            )

        if self.validate and not self.default_file:
            raise ValueError(
                'Cannot use "validate" without specifying a ' "default file."
            )

    def read_default_config(self):
        """Read the default config file.

        :raises DefaultConfigValidationError: There was a validation error with
                                              the *default* file.
        """
        if self.validate:
            self.default_config = ConfigObj(
                configspec=self.default_file,
                list_values=False,
                _inspec=True,
                encoding="utf8",
            )
            # ConfigObj does not set the encoding on the configspec.
            self.default_config.configspec.encoding = "utf8"

            valid = self.default_config.validate(
                Validator(), copy=True, preserve_errors=True
            )
            if valid is not True:
                for name, section in valid.items():
                    if section is True:
                        continue
                    for key, value in section.items():
                        if isinstance(value, ValidateError):
                            raise DefaultConfigValidationError(
                                'section [{}], key "{}": {}'.format(name, key, value)
                            )
        elif self.default_file:
            self.default_config, _ = self.read_config_file(self.default_file)

        self.update(self.default_config)

    def read(self):
        """Read the default, additional, system, and user config files.

        :raises DefaultConfigValidationError: There was a validation error with
                                              the *default* file.
        """
        if self.default_file:
            self.read_default_config()
        return self.read_config_files(self.all_config_files())

    def user_config_file(self):
        """Get the absolute path to the user config file."""
        return os.path.join(
            get_user_config_dir(self.app_name, self.app_author), self.filename
        )

    def system_config_files(self):
        """Get a list of absolute paths to the system config files."""
        return [
            os.path.join(f, self.filename)
            for f in get_system_config_dirs(self.app_name, self.app_author)
        ]

    def additional_files(self):
        """Get a list of absolute paths to the additional config files."""
        return [os.path.join(f, self.filename) for f in self.additional_dirs]

    def all_config_files(self):
        """Get a list of absolute paths to all the config files."""
        return (
            self.additional_files()
            + self.system_config_files()
            + [self.user_config_file()]
        )

    def write_default_config(self, overwrite=False):
        """Write the default config to the user's config file.

        :param bool overwrite: Write over an existing config if it exists.
        """
        destination = self.user_config_file()
        if not overwrite and os.path.exists(destination):
            return

        with io.open(destination, mode="wb") as f:
            self.default_config.write(f)

    def write(self, outfile=None, section=None):
        """Write the current config to a file (defaults to user config).

        :param str outfile: The path to the file to write to.
        :param None/str section: The config section to write, or :data:`None`
                                 to write the entire config.
        """
        with io.open(outfile or self.user_config_file(), "wb") as f:
            self.data.write(outfile=f, section=section)

    def read_config_file(self, f):
        """Read a config file *f*.

        :param str f: The path to a file to read.
        """
        configspec = self.default_file if self.validate else None
        try:
            config = ConfigObj(
                infile=f, configspec=configspec, interpolation=False, encoding="utf8"
            )
            # ConfigObj does not set the encoding on the configspec.
            if config.configspec is not None:
                config.configspec.encoding = "utf8"
        except ConfigObjError as e:
            logger.warning(
                "Unable to parse line {} of config file {}".format(e.line_number, f)
            )
            config = e.config

        valid = True
        if self.validate:
            valid = config.validate(Validator(), preserve_errors=True, copy=True)
        if bool(config):
            self.config_filenames.append(config.filename)

        return config, valid

    def read_config_files(self, files):
        """Read a list of config files.

        :param iterable files: An iterable (e.g. list) of files to read.
        """
        errors = {}
        for _file in files:
            config, valid = self.read_config_file(_file)
            self.update(config)
            if valid is not True:
                errors[_file] = valid
        return errors or True


def get_user_config_dir(app_name, app_author, roaming=True, force_xdg=True):
    """Returns the config folder for the application.  The default behavior
    is to return whatever is most appropriate for the operating system.

    For an example application called ``"My App"`` by ``"Acme"``,
    something like the following folders could be returned:

    macOS (non-XDG):
      ``~/Library/Application Support/My App``
    Mac OS X (XDG):
      ``~/.config/my-app``
    Unix:
      ``~/.config/my-app``
    Windows 7 (roaming):
      ``C:\\Users\\<user>\\AppData\\Roaming\\Acme\\My App``
    Windows 7 (not roaming):
      ``C:\\Users\\<user>\\AppData\\Local\\Acme\\My App``

    :param app_name: the application name. This should be properly capitalized
                     and can contain whitespace.
    :param app_author: The app author's name (or company). This should be
                       properly capitalized and can contain whitespace.
    :param roaming: controls if the folder should be roaming or not on Windows.
                    Has no effect on non-Windows systems.
    :param force_xdg: if this is set to `True`, then on macOS the XDG Base
                      Directory Specification will be followed. Has no effect
                      on non-macOS systems.

    """
    if WIN:
        key = "APPDATA" if roaming else "LOCALAPPDATA"
        folder = os.path.expanduser(os.environ.get(key, "~"))
        return os.path.join(folder, app_author, app_name)
    if MAC and not force_xdg:
        return os.path.join(
            os.path.expanduser("~/Library/Application Support"), app_name
        )
    return os.path.join(
        os.path.expanduser(os.environ.get("XDG_CONFIG_HOME", "~/.config")),
        _pathify(app_name),
    )


def get_system_config_dirs(app_name, app_author, force_xdg=True):
    r"""Returns a list of system-wide config folders for the application.

    For an example application called ``"My App"`` by ``"Acme"``,
    something like the following folders could be returned:

    macOS (non-XDG):
      ``['/Library/Application Support/My App']``
    Mac OS X (XDG):
      ``['/etc/xdg/my-app']``
    Unix:
      ``['/etc/xdg/my-app']``
    Windows 7:
      ``['C:\ProgramData\Acme\My App']``

    :param app_name: the application name. This should be properly capitalized
                     and can contain whitespace.
    :param app_author: The app author's name (or company). This should be
                       properly capitalized and can contain whitespace.
    :param force_xdg: if this is set to `True`, then on macOS the XDG Base
                      Directory Specification will be followed. Has no effect
                      on non-macOS systems.

    """
    if WIN:
        folder = os.environ.get("PROGRAMDATA")
        return [os.path.join(folder, app_author, app_name)]
    if MAC and not force_xdg:
        return [os.path.join("/Library/Application Support", app_name)]
    dirs = os.environ.get("XDG_CONFIG_DIRS", "/etc/xdg")
    paths = [os.path.expanduser(x) for x in dirs.split(os.pathsep)]
    return [os.path.join(d, _pathify(app_name)) for d in paths]


def _pathify(s):
    """Convert spaces to hyphens and lowercase a string."""
    return "-".join(s.split()).lower()



# -*- coding: utf-8 -*-
"""OS and Python compatibility support."""

from decimal import Decimal
from types import SimpleNamespace
import sys

PY2 = sys.version_info[0] == 2
WIN = sys.platform.startswith("win")
MAC = sys.platform == "darwin"


if PY2:
    text_type = unicode
    binary_type = str
    long_type = long
    int_types = (int, long)

    from UserDict import UserDict
    from backports import csv

    from StringIO import StringIO
    from itertools import izip_longest as zip_longest
else:
    text_type = str
    binary_type = bytes
    long_type = int
    int_types = (int,)

    from collections import UserDict
    import csv
    from io import StringIO
    from itertools import zip_longest


HAS_PYGMENTS = True
try:
    from pygments.token import Token
    from pygments.formatters.terminal256 import (
        TerminalTrueColorFormatter,
        Terminal256Formatter,
    )
except ImportError:
    HAS_PYGMENTS = False
    Terminal256Formatter = None
    TerminalTrueColorFormatter = None
    Token = SimpleNamespace()
    Token.Output = SimpleNamespace()
    Token.Output.Header = None
    Token.Output.OddRow = None
    Token.Output.EvenRow = None
    Token.Output.Null = None
    Token.Output.TableSeparator = None
    Token.Results = SimpleNamespace()
    Token.Results.Header = None
    Token.Results.OddRow = None
    Token.Results.EvenRow = None


float_types = (float, Decimal)



# -*- coding: utf-8 -*-
"""A tsv data output adapter"""

from __future__ import unicode_literals

from .preprocessors import bytes_to_string, override_missing_value, convert_to_string
from itertools import chain
from cli_helpers.utils import replace

supported_formats = ("tsv", "tsv_noheader")
preprocessors = (override_missing_value, bytes_to_string, convert_to_string)


def adapter(data, headers, table_format="tsv", **kwargs):
    """Wrap the formatting inside a function for TabularOutputFormatter."""
    if table_format == "tsv":
        for row in chain((headers,), data):
            yield "\t".join((replace(r, (("\n", r"\n"), ("\t", r"\t"))) for r in row))
    elif table_format == "tsv_noheader":
        for row in data:
            yield "\t".join((replace(r, (("\n", r"\n"), ("\t", r"\t"))) for r in row))
    else:
        raise ValueError(f"Invalid table_format specified: {table_format}.")
